HAL - общие абстракции драйверов железа
=============================================================
библиотека АПИ устройств, и АПИ ОС.

платформы
--------------------------------------------------------------------
- `MDR32F9Qx/` - [порт для SPL 1886ВЕ94Т](MDR32F9Qx/README.md)
- `STM32*/` - порт для SPL STM32F4 - не реализован, не отлажен

Портирование
=============================================================
адаптация SPL железа
--------------------------------------------------------------------
библиотека HAL спроектирована с зацелом отделения функциональной части от 
аппаратной реализции.

хедеры вида `xxx_hal.hpp` - объявляют интерфейс драйвера, и имеют общий 
аппаратно-независимый код. Реализации драйвера под конкретную платформу создаются 
в папках платформы. 

сейчас реализован порт SPL Миландра для контроллеров 1886ВЕxx - [`MDR32F9Qx/`](MDR32F9Qx/README.md)

хедеры интерфейсов драйверов HAL тоже используют SPL платформы проекта, но не напрямую 
а через прокси-хедеры вида `mcu_xxx.h`. Задача их - включить нужный хедер SPL,
и адаптировать его вызовы и названия к общему виду, используемому HAL.

так же порт реализует обязательный хедер `gpio.h` - управление и настройку пинов контроллера

Для порта на другой SPL надо завести каталог платформы и, как минимум, 
написать новые `gpio.h` и `mcu_xxx.h`.

И сделать версии драйверов HAL унаследованные от интерфейсов `xxx_hal.hpp`. 
Эти реализации должны обеспечить настройку и иницализацию железа под драйвер.

адаптация к OS
--------------------------------------------------------------------
Драйвера, требующие конкурентного доступа от разных процессов, или работы с системными
часами и таймаутами используют ресурсы ОС. 
HAL использует ресурсы ОС через хедеры обертки-порты:
- `OsSync.h` - порт примитивов синхронизации ОС: 
     - *критические секци*
     - *семафор*
     - *мутекс*
     - *сигнал*
     - нитки, их управление
     - sleep

  собственно HAL использует *выделенное*, остальные ресурсы используются кодом проекта.

- `OsTime.h` - порт примитивов работы с временем: таймеры и, таймауты
- `os_isr.h` - порт управления преерываниями: вкл/выкл прерывания
- `system.h` - все остальное. Сейчас дает частоту процессора, и пустое ожидание события. 
- `halos_types.h` - хедер связывающий простые типы HAL - результаты, ошибки,
        с константами и типами ОС.

Структура
=============================================================

Общие файлы
--------------------------------------------------------------------
- `hal_types.h` - удобные типы для использования в ХАЛ.

- `halos_types.h` - хедер-образец. порты ОС, по мотивам этого хедера, адаптирую
  типы ХАЛ с типами ОС

- `hal_device.h` - перечисление интерфейсов различных устройств общего вида.
 
- `porting/` - каталог портов АПИ разных [ОС](README-OS.md).

адаптация к компилятору
-------------------------------------------------------------------

хедеры конфигурации
-------------------------------------------------------------------
- `project-conf.h` - хедер-образец. по этому образцу приложение делает этот 
  же хедер, с настройками ХАЛ под проект.

библиотеки
--------------------------------------------------------------------
- `hal_ringbuffer` - шаблоны кольцевой буффера. используются в буферах 
  ввода/вывода - UART, SLIP 
- `slip-filter` - фильтр протокола SLIP прямо на буффере HAL_RingBuffer.

драйвера
-------------------------------------------------------------------
файлы `*_hal` - описание интерфейса соотвествующего устройства. 

- uart_hal - драйвер УАРТа, с циклическим буфером на приемнике и трансмитере.
- slip_hal - прокси-фильтр УАРТА фреймов протокола SLIP  
- switch_hal - переключатель набора пинов контроллера, с одним активным пином.
- timer_hal - базовый таймер/счетчик/ШИМ. 
- ssp\_hal - драйвер SPI, и его прокси SPIIO реализующий `IO_Device` примитивы
- flash_hal - драйверы флеш-памяти, флеш на SPI, и банка флешек.
- flash/    - каталог реализаций драйверов разных флешек
    - 1636RR52 - драйвер spi-флешки 1636РР52  
