﻿/*
 *      Author: alexrayne <alexraynepe196@gmail.com>
  ------------------------------------------------------------------------
    Copyright (c) alexrayne

   All rights reserved.
   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are met:
   - Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
   - Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
   - Neither the name of ARM nor the names of its contributors may be used
     to endorse or promote products derived from this software without
     specific prior written permission.
   *
   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
   ARE DISCLAIMED. IN NO EVENT SHALL COPYRIGHT HOLDERS AND CONTRIBUTORS BE
   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
   POSSIBILITY OF SUCH DAMAGE. *
*/
#include <ctype.h>
#include <cstring>
#include "cli.hpp"

using namespace cli;

//virtual
int CLICommand::cli_process(CLI_shell* shell, const_line cmd){
    const_line line = cli::after_name(cmd, cmd_name);
    return line - cmd;
};

bool CLICommand::registerCmd(CLI_shell* shell){
  if (shell == NULL)
    return cli::registerCmd(this);
  else
    return shell->registerCmd(this);
}

//virtual
void CLICommand::cli_help(CLI_shell* shell){
    shell->printf( "%s \n", cmd_name );
}


bool CLICommand::is_cmd_starts(const_line line){
  return str_starts(line, cmd_name) > 0;
}

CLICommand::const_line CLICommand::after_cmd(const_line cmd){
    return cli::after_name(cmd, cmd_name);
}



static const unsigned cli_cmd_reserve = 4;
CLI_commands::CLI_commands(){
  reserve(cli_cmd_reserve);
};

CLI_commands::CLI_commands(unsigned _reserves){
  reserve(_reserves);
}

bool CLI_commands::append(CLICommand *cmd){
  push_back(cmd);
  return back() == cmd;
}

void CLI_commands::help(CLI_shell* shell){
  shell->puts("avail commands:\n");
  for(iterator _it = this->begin(); _it != this->end(); ++_it){
    CLICommand *cmd = *_it;
    if (cmd->cmd_name == NULL)
        continue;
    shell->puts(cmd->cmd_name);
  }
}

int CLI_commands::cli_process(CLI_shell* shell, CLI_commands::const_line line){
    assert(line != NULL);
  for(iterator _it = this->begin(); _it != this->end(); ++_it){
    CLICommand *cmd = *_it;
    int res = cmd->cli_process(shell, line);
    if (res != 0)
      return res;
  }
  return 0;
}


CLICommand* CLI_commands::lookup(CLI_commands::const_line cmd){
  if (cmd == NULL)
      return NULL;
  const_line name = cmd + space_len(cmd, ~0);
  for(iterator _it = this->begin(); _it != this->end(); ++_it){
    CLICommand *cmd = *_it;
    if (cmd->is_cmd_starts(name))
      return cmd;
  }
  return NULL;
}




cli::CLI_Help::CLI_Help(CLI_commands* dst)
: inherited("help")
{
  dst->append(this);
}

//virtual
int cli::CLI_Help::cli_process(CLI_shell* shell, CLICommand::const_line line)
{
    assert(line != NULL);
  int len = name_starts(line, cmd_name);
  if (len > 0){
    const_line s = line + len;
    s += space_len(s, ~0);
    len = is_eol(s);
    s += len;
    if (len < 0){
      CLICommand* cmd = shell->lookupCmd(s);
      if (cmd != NULL){
        s += name_starts(s, cmd->cmd_name);
        cmd->cli_help(shell);
      }
      else
        shell->help();
    }
    else {
      shell->help();
    }
    return s - line;
  }
  return 0;
}
//* печатает хелп команды в шелл
//virtual
void cli::CLI_Help::cli_help(CLI_shell* shell){
  shell->help();
}

const char* const cli::names_onoff[] = { "off", "on", NULL};
int cli::mutch_onoff(const char* __restrict line
                , const char** __restrict lend
                )
{
    return mutch_names(line, lend, names_onoff);
}


cli::CLI_OnOffCommand::CLI_OnOffCommand():inherited(){;}

cli::CLI_OnOffCommand::CLI_OnOffCommand(const_cmd _cmd)
:inherited(_cmd)
{;}


//virtual
int cli::CLI_OnOffCommand::cli_process(CLI_shell* shell, const_line cmd){
    int ok = CLICommand::cli_process(shell, cmd);
    if (ok <= 0)
        return 0;

    const_line cmdval = cmd + ok;
    bool value = false;
    const_line line = after_name(cmdval, "on");
    if (line != NULL)
        value = true;
    else {
        line = after_name(cmdval, "off");
        if (line == NULL)
            return 0;
    }
    cli_onoff(value);
    return line-cmd;
}
//* печатает хелп команды в шелл
//virtual
void cli::CLI_OnOffCommand::cli_help(CLI_shell* shell){
    shell->puts("on|off \t - enable/disable\n");
}






CLI_commands CLI_shell::cmds(cli_cmd_reserve);
CLI_Help     CLI_shell::help_cmd(&CLI_shell::cmds);

bool cli::registerCmd(CLICommand *cmd){
  return CLI_shell::cmds.append(cmd);
}

int cli::space_len(const char* s, unsigned len){
    if (s == NULL)
        return 0;
  const char* tmp = s;
  for(; len > 0; len--, tmp++){
    char c = *tmp;
    if (c == 0)
      break;
    if (!isspace(c))
      break;
  }
  return tmp - s;
}

int cli::word_len(const char* s, unsigned len){
    if (s == NULL)
        return 0;
  const char* tmp = s;
  for(; len > 0; len--, tmp++){
    char c = *tmp;
    if (c == 0)
      break;
    if (isalnum(c)||ispunct(c))
      continue;
    break;
  }
  return tmp - s;
}

//* word_len + check that after word is space or EOL
int cli::name_len(const char* s, unsigned len){
    if (s == NULL)
        return 0;
  int res = word_len(s, len);
  if (res >= len)
    return res;
  if (is_eol(s+res) >= 0)
    return res;
  if (isspace(s[res]))
    return res;
  return -1;
}

//* look to EOL or Z
int cli::line_len(const char* s, unsigned len){
    if (s == NULL)
        return 0;
  const char* tmp = s;
  for(; len > 0; len--, tmp++){
    int el = is_eol(tmp);
    if (el < 0)
      continue;
    tmp += el;
    break;
  }
  return tmp - s;
}

//* \return < 0 - no eol
//* \returm = 0 - Z ended
//* \return >0  - len of EOL token
int cli::is_eol(const char* s){
    if(s == NULL)
        return 0;
  char c = *s;
  if (c == '\0')
    return 0;
  if (c == '\n')
    return 1;
  return -1;
}

//* suggests that <name> shorter vs <line>
//* \return = 0 - <name> not starts <line>
//* \return > 0 - len of <name> when one completely starts <line>
//* \return < 0 - len of same prefix in <line> and <name>
int cli::str_starts(const char* __restrict line, const char* __restrict name){
  if ((name == NULL) || (line == NULL))
    return 0;
  const char* a = line;
  const char* b = name;
  for ( ; *a == *b ; a++, b++){
    if (*a == 0) break;
  }
  if (*b == 0)
    return b - name;
  return -(a - line);
}

//* same as str_starts, ensures that mutched prefix <name> in <line>
//*    finishes by EOL, or space
int cli::name_starts(const char* __restrict line, const char* __restrict name){
  if ((name == NULL) || (line == NULL))
    return 0;
  int slen = space_len(line, ~0);
  line += slen;
  int res = str_starts(line, name);
  if (res > 0){
    char c = line[res];
    if ( (c == 0)
      || isspace(c)
      || (c == '\n')
      )
      return res+slen;
    else
      return -res-slen;
  }
  return 0;
}

const char* cli::after_name(const char* __restrict line
                            , const char* __restrict name)
{
    int len = cli::name_starts(line, name);
    if ( len <= 0 )
      return 0;
    line += len;
    line += cli::space_len(line, ~0);
    return line;
}

//* \return index of mutched in <names>
//*         , <lend> = position in <line> after mutched
//* \return  = -1 -  if fail mutches
int cli::mutch_names(const char* __restrict line
                    , const char** __restrict lend
                , const char* const * __restrict names)
{
  for (unsigned i = 0; *names != NULL; i++, names++ ){
    const char* name = *names;
    int len = name_starts(line, name);
    if (len > 0){
      *lend = line + len;
      return i;
    }
  }
  return -1;
}

// \return > 0 - anount of accepted chars, x = parsed value
int cli::take_number(long& x, const char* s){
    const char* line = s;
    line += space_len(line, ~0);
    if (is_eol(line) >= 0)
        return 0;
    if (*line == '0'){
        if (strchr("xXbBoO", line[1]) != NULL){
            x = strtoul(line, (char**)&line, 0);
        }
        else
            x = strtol(line, (char**)&line, 0);
    }
    else if (*line == '$'){
        x = strtoul(line+1, (char**)&line, 16);
    }
    else {
        x = strtol(line, (char**)&line, 0);
    }
    return line-s;
}

// \return > 0 - anount of accepted chars, x = parsed value
//           s - s updates to new position after number
const char* cli::pass_number(long& x, const char* s){
    int len = take_number(x, s);
    if (len > 0)
        return s + len;
    return NULL;
}

const char* cli::pass_number(char mark, long& x, const char* s){
    const char* line = s;

    line += space_len(line, ~0);
    if (is_eol(line) >= 0)
        return 0;

    if (*line++ != mark)
        return NULL;

    int len = take_number(x, line);
    if (len > 0)
        return line + len;

    return NULL;
}

const char* cli::pass_number( char markx, long& x
                        ,char marky, long& y, const char* s)
{
    const char* line = s;
    line += space_len(line, ~0);
    if (is_eol(line) >= 0)
        return 0;

    if (*line++ != markx)
        return NULL;

    int len = take_number(x, line);
    if (len <= 0)
        return NULL;

    line += len;
    if (*line++ != marky)
        return NULL;

    len = take_number(y, line);
    if (len <= 0)
        return NULL;

    return line + len;
}



CLI_shell::CLI_shell()
: mycmds(cli_cmd_reserve)
{
  init();
};

void CLI_shell::init(){
  cmdline.reserve(cmdlimit);
  clear();
}

bool CLI_shell::registerCmd(CLICommand *cmd){
  return mycmds.append(cmd);
}

CLICommand* CLI_shell::lookupCmd(CLICommand::const_line cmd){
  CLICommand* res = mycmds.lookup(cmd);
  if (res != NULL)
    return res;
  res = cmds.lookup(cmd);
  return res;
}

void CLI_shell::help(){
  mycmds.help(this);
  cmds.help(this);
}

void CLI_shell::clear(){
  cmdline.resize(1);
  cmdline.front() = '\0';
  cursor = &cmdline.back();
}

//* продвигает курсор парсера в командной строке
const char* CLI_shell::acceptChars(int len){
  if (cursor_len >= len){
    cursor     += len;
    cursor_len -= len;
  }
  else {
    cursor     += cursor_len;
    cursor_len = 0;
    if (cursor == cmdline.end())
      clear();
  }
  return cursor;
}

const char* CLI_shell::acceptChars(const char* line){
    assert((line > cursor));
  int len = line - cursor;
  if (len < 0)
    return NULL;
  return acceptChars(len);
}


//* добавляет символ в строку, и выполняет команду на LN в контексте вызвавшего
void CLI_shell::addChar(u8 ch){
  cmdline.back() = ch;
  cmdline.push_back('\0');
};

//* добавляет символ в начало строки
void CLI_shell::unwindChar(u8 ch){
    cursor--;
    *(char*)(cursor) = ch;
    cursor_len++;
}

//* исполняет набранную в cmdline строку
void CLI_shell::processCmd(){
    if (cmdline.size() <= 0)
        return;

  processLine( cmdline.begin(), cmdline.size());
  acceptChars(cli::space_len(cursor,cursor_len));
}

void CLI_shell::dropProcessed( void ){
    if (cursor_len == 0){
      clear();
      return;
    }
    int accepted = cmdline.size() - cursor_len;
    cmdline.drop(cmdline.begin(), accepted);
    cursor = cmdline.begin();
    cursor_len = cmdline.size();
}

void CLI_shell::processLine(const char* line, unsigned len){
    assert( (line != NULL) && (len > 0) );
  cursor = line;
  cursor_len = len;
  for (acceptChars(space_len(cursor, cursor_len))
      ; cursor_len > 0
      ; acceptChars(space_len(cursor, cursor_len)) )
  {
    int linelen = line_len(cursor, cursor_len);
    if (linelen <= 0)
        continue;
    if (is_eol(cursor) >= 0)
        continue;
#ifdef LOGAPI
    debug.log(vINFO, "cli:cmd:%.*s\n", linelen, cursor);
    debug.flush();
#endif
    int parsed = 0;
    parsed = mycmds.cli_process(this, cursor);
    if (parsed > 0){
      cursor = acceptChars(parsed);
      continue;
    }
    else if (parsed < 0)
      break;

    parsed = cmds.cli_process(this, cursor);
    if (parsed > 0){
      cursor = acceptChars(parsed);
      continue;
    }
    else if (parsed < 0)
      break;
    else {
      int LL = cli::line_len(cursor, cursor_len);
      int show_LL = (LL < 10)? LL : 10;
      printf("uncknown cli command:%.*s...\n", show_LL, cursor);
      acceptChars(LL);
      break;
    }
  }
}

//virtual
int CLI_shell::getLine(char* line, unsigned limit, char eol){
  assert((line != NULL) && (limit > 0));
  if (cursor_len <= 0){
    *line = '\0';
    return 0;
  }
  const char* src = cursor;
  char* dst = line;
  limit--;
  for ( ; limit > 0; limit--){
    char c = *src++;
    *dst++ = c;
    if ( (c == eol) || (c == '\0') )
    {
        break;
    }
  }
  *dst = '\0';
  return dst - line;
}

//virtual
int CLI_shell::puts(const char* s){
  return 0;
}

//virtual
bool CLI_shell::ask_yes(void){
  if (cursor_len <= 0)
    return false;
  acceptChars( cli::space_len(cursor, cursor_len) );
  char c = *cursor;
  switch (c){
    case 0: return false;
    case 'y':
    case 'Y':
      acceptChars(1);
      return true;
    default:
      acceptChars(1);
      return false;
  }
}



