﻿/*
 *      Author: alexrayne <alexraynepe196@gmail.com>
  ------------------------------------------------------------------------
    Copyright (c) alexrayne

   All rights reserved.
   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are met:
   - Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
   - Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
   - Neither the name of ARM nor the names of its contributors may be used
     to endorse or promote products derived from this software without
     specific prior written permission.
   *
   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
   ARE DISCLAIMED. IN NO EVENT SHALL COPYRIGHT HOLDERS AND CONTRIBUTORS BE
   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
   POSSIBILITY OF SUCH DAMAGE. *
*/
#ifndef V12_CLI_H_
#define V12_CLI_H_

#include "intfaces.h"
//#include <vector>

#define BUFFERLEN 64
#define MAXTOKEN 5+1
#define ENDL '\r'
#define MAX_CMDS 3
#define MAX_FUNS 3

class CLI_shell;

class CLICommand{
public:
    typedef const_string const_cmd;
    typedef char* cmd_line;
    typedef const char* const_line;
    const_cmd cmd_name;

    CLICommand():cmd_name(NULL){};
    CLICommand(const_cmd _cmd):cmd_name(_cmd){};
    ~CLICommand(){
    };
    //* \return > 0- длинна принятой строки
    //*         ==0 - команда не распознана
    //*         <0  - ошибка
    virtual int cli_process(CLI_shell* shell, const_line line){
      return 0;
    };
    //* печатает хелп команды в шелл
    virtual void cli_help(CLI_shell* shell){
    };

    bool is_cmd_starts(const_line line);

public:
    bool registerCmd(CLI_shell* shell = NULL);
};


typedef std::vector< CLICommand*, SystemHeap< CLICommand* > > CLIvector;

class CLI_commands
    : public CLIvector
{
  public:
    typedef CLIvector inherited;
    typedef CLICommand::cmd_line cmd_line;
    typedef CLICommand::const_line const_line;

    CLI_commands();
    CLI_commands(unsigned reserve);
    bool append(CLICommand *cmd);
    CLICommand* lookup(const_line cmd);

    void help(CLI_shell* shell);
    int cli_process(CLI_shell* shell, const_line line);
};

namespace cli {
//* регистрирует глобальную команду в CLI_shell::cmds
bool registerCmd(CLICommand *cmd);

int space_len(const char* s, unsigned len);
int word_len(const char* s, unsigned len);
//* word_len + check that after word is space or EOL
//* \return -1 if ccheck fails
int name_len(const char* s, unsigned len);
//* look to EOL or Z
int line_len(const char* s, unsigned len);
//* \return < 0 - no eol
//* \returm = 0 - Z ended
//* \return >0  - len of EOL token
int is_eol(const char* s);

//* suggests that <name> shorter vs <line>
//* \return = 0 - <name> not starts <line>
//* \return > 0 - len of <name> when one completely starts <line>
//* \return < 0 - len of same prefix in <line> and <name>
int str_starts(const char* __restrict line, const char* __restrict name);
//* same as str_starts, ensures that mutched prefix <name> in <line>
//*    finishes by EOL, or space, skip spaces before name
int name_starts(const char* __restrict line, const char* __restrict name);

//* \return index of mutched in <names>
//*         , <lend> = position in <line> after mutched
//* \return  = -1 -  if fail mutches
int mutch_names(const char* __restrict line
                , const char** __restrict lend
                , const char* const * __restrict names);


class CLI_Help : public CLICommand{
  public:
    typedef CLICommand inherited;

    CLI_Help(CLI_commands* dst);

    virtual int cli_process(CLI_shell* shell, const_line line);
    //* печатает хелп команды в шелл
    virtual void cli_help(CLI_shell* shell);
};

}; //namespace cli

class CLI_shell
  : public virtual StdPrint
{
public:
    static CLI_commands cmds;
    static cli::CLI_Help help_cmd;

public:
    CLI_shell();
    void init();

    CLI_commands mycmds;
    bool registerCmd(CLICommand *cmd);
    CLICommand* lookupCmd(CLICommand::const_line cmd);
public:
    typedef std::vector <char, SystemHeap <char> > chars;
    typedef std::vector <chars::iterator> positions;
    static const unsigned cmdlimit = 164;
    chars cmdline;      // cmdline buffer
    const char* cursor;
    unsigned    cursor_len;

    void help();
    void clear();
    //* исполняет набранную в cmdline строку
    void processCmd();
    //* добавляет символ в строку, и выполняет команду на LN в контексте вызвавшего
    void addChar(u8 ch);
    //* добавляет символ в начало строки
    void unwindChar(u8 ch);

    void processLine(const char* line, unsigned len);

    //* продвигает курсор парсера в командной строке
    const char* acceptChars(int len);
    const char* acceptChars(const char* line);
    virtual int getLine(char* line, unsigned limit, char eol = '\n');
    virtual int puts(const char* s);
    virtual bool ask_yes(void);

    virtual void on_fail(const char* s){};
};

typedef StdOUT_PrintBuffer<512> CLIShellPrinter;

class Device_CLI_shell
    : public CLI_shell
    , public StdOUT_ProxyPrint
    //,  public Device_Observer
{
public:
    typedef CLI_shell inherited;
    Device_CLI_shell();
    Device_CLI_shell(StdIO_Device *device);
//    virtual void handleEvent(StdIO_Device_Observable *observable){this->addChar(observable->get_char());}
//    virtual void handleEvent(Observer_Meta *observable){this->handleEvent(((StdIO_Device_Observable * )observable));}
protected:
    CLIShellPrinter     printer;
    StdIO_Device*       device;
public:
    static const unsigned get_poll_toms = 100;

    void init(StdIO_Device *device);
    StdIO_Device* io(){return device;};
    StdPrint&     out(){return printer;};
    void execute(const char* prompt = ">");

    void addChar(u8 ch);
    virtual int getLine(char* line, unsigned limit, char eol = '\n');
    //* readline from io
    int readLine(char* line, unsigned limit, char eol = '\n');
    virtual int puts(const char* s);

    virtual void on_fail(const char* s);
    virtual bool ask_yes(void);
};



#endif //V12_CLI_H_
