/*
 * hal_shell.cpp
 *
 *  Created on: 19 мар. 2019 г.
 *      Author: Lityagin Aleksandr
 *      Author: alexrayne <alexraynepe196@gmail.com>
  ------------------------------------------------------------------------
    Copyright (c) alexrayne

   All rights reserved.
   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are met:
   - Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
   - Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
   - Neither the name of ARM nor the names of its contributors may be used
     to endorse or promote products derived from this software without
     specific prior written permission.
   *
   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
   ARE DISCLAIMED. IN NO EVENT SHALL COPYRIGHT HOLDERS AND CONTRIBUTORS BE
   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
   POSSIBILITY OF SUCH DAMAGE. *
 */

#include "hal_shell.hpp"
#include <OsSync.h>

Device_CLI_shell::Device_CLI_shell()
:inherited()
,StdOUT_ProxyPrint(&printer)
{
}

Device_CLI_shell::Device_CLI_shell(StdIO_Device* con)
:inherited()
,StdOUT_ProxyPrint(&printer)
{
  init(con);
}

void Device_CLI_shell::addChar(u8 ch){
  inherited::addChar(ch);
}

//virtual
int Device_CLI_shell::getLine(char* line, unsigned limit, char eol){
    assert((line != NULL) && (limit > 0));
  char* dst = line;
  int len = inherited::getLine(line, limit, eol);
  if (len > 0){
    dst += len;
    limit -= len;
    if ((limit == 0) || (dst[-1] == eol))
      return len;
  }
  len += readLine(dst, limit, eol);
  return len;
}

int Device_CLI_shell::readLine(char* line, unsigned limit, char eol){
    assert((line != NULL) && (limit > 0));
  char* dst = line;
  int len = 0;
  limit--;
  for (; limit >0 ; limit--){
     *dst = '\0';
     int tmp = device->get_char();
     while (tmp < 0){
       if (device->get_wait( OSTicksMS(get_poll_toms) ) < 0){
         return len;
       };
       tmp = device->get_char();
     }
     char c = tmp;
     device->putChar(c);    // echo
     *dst = c;
     if (c == '\0')
       break;
     dst    += 1;
     len    += 1;
     if (c == eol)
       break;
  }
  *dst = '\0';
  return len;
}
//virtual
int Device_CLI_shell::puts(const char* s){
  return device->puts(s);
}

void Device_CLI_shell::init(StdIO_Device * _device){
  device = _device;
  printer.init(device);
}

void Device_CLI_shell::execute(const char* prompt){
    while(device){
      cmdline.resize(cmdlimit);
      if (prompt)
        io()->puts(prompt);
      int len = getLine(cmdline.begin(), cmdlimit, '\n');
      if (len > 0){
        cmdline.resize(len);
        cursor_len = len;
        cursor = cmdline.begin();
        processCmd();
        dropProcessed();
      }
      else
        sleep_ms(get_poll_toms);
    }
}

//virtual
void Device_CLI_shell::on_fail(const char* s){
    assert(s != NULL);
  char buf[64];
  //int len =
  int LL = cli::line_len(s, ~0);
  int show_LL = (LL < 10)? LL : 10;
  snprintf(buf, sizeof(buf),"failed parse command at:%.*s...\n"
           , show_LL, s);
  puts(buf);
}

//virtual
bool Device_CLI_shell::ask_yes(void){
  char tmp[2];
  int len = readLine(tmp, 2, 0);
  if (len != 1)
    return false;
  //char c = device->get_char();;
  char c = tmp[0];
  switch (c){
    case 0: return false;
    case 'y':
    case 'Y':
      acceptChars(1);
      return true;
    default:
      acceptChars(1);
      return false;
  }
}


