/*
 * observe_pattern.hpp
 *
 *  Created on: 18 мар. 2019 г.
 *      Author: alexrayne <alexraynepe196@gmail.com>
  ------------------------------------------------------------------------
    Copyright (c) alexrayne

   All rights reserved.
   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are met:
   - Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
   - Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
   - Neither the name of ARM nor the names of its contributors may be used
     to endorse or promote products derived from this software without
     specific prior written permission.
   *
   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
   ARE DISCLAIMED. IN NO EVENT SHALL COPYRIGHT HOLDERS AND CONTRIBUTORS BE
   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
   POSSIBILITY OF SUCH DAMAGE. *
 */

#ifndef LIB_OBSERVE_PATTERN_HPP_
#define LIB_OBSERVE_PATTERN_HPP_



template <typename T, typename Tstamp = sec_t>
class Observer {
public:
    typedef T       value_t;
    typedef Tstamp  stamp_t;
    virtual void process(T chunk, Tstamp ts) { };
    virtual void update(Tstamp ts) { };
private:

};



template <typename T, int observableSize>
class staticObservable {
public:
  typedef Observer<T> IObserver;
  static const unsigned observersLimit = observableSize;
  void init () {
    observableHead = 0;
  }

  virtual void addObserver(IObserver* observer) {
    if(observableHead < observersLimit)
      observers[observableHead++] = observer;
  }

  virtual void process_observers(T chunk, sec_t ts) {
    for (int i=0; i < observableHead; i++){
      observers[i]->process(chunk, ts);
    }
  }

  virtual void update_observers(sec_t ts) {
    for (int i=0; i < observableHead; i++){
      observers[i]->update(ts);
    }
  }

protected:
    IObserver* observers[observersLimit];
    int observableHead;
};



class Observer_Meta{
    //TODO Remove this '''kostil'''
};

template<unsigned observers_limit>
class Observables
    : public Observer_Meta
    , public staticObservable<Observer_Meta*, observers_limit>
{
};

typedef class Observer<Observer_Meta*> Observer;
typedef class Observables<4>    Observable;



#endif /* LIB_OBSERVE_PATTERN_HPP_ */
